/*
 * Simulator of microcontrollers (uc51.cc)
 *
 * Copyright (C) 1999,99 Drotos Daniel, Talker Bt.
 * 
 * To contact author send email to drdani@mazsola.iit.uni-miskolc.hu
 *
 */

/* This file is part of microcontroller simulator: ucsim.

UCSIM is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

UCSIM is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with UCSIM; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA. */
/*@1@*/

#include "ddconfig.h"

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <termios.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#if FD_HEADER_OK
# include HEADER_FD
#endif
#include "i_string.h"

// sim
#include "optioncl.h"

// local
#include "uc51cl.h"
#include "glob.h"
#include "regs51.h"


/*
 * Making a new micro-controller and reset it
 */

t_uc51::t_uc51(int Itype, int Itech, class cl_sim *asim):
  cl_uc(asim)
{
  int i;
  struct termios tattr;
  
  type= Itype;
  technology= Itech;

  debug= asim->get_iarg('V', 0);
  stop_at_it= FALSE;
  options->add(new cl_bool_opt(&debug, "verbose", "Verbose flag."));
  options->add(new cl_bool_opt(&stop_at_it, "stopit",
			       "Stop if interrupt accepted."));
  options->add(new cl_cons_debug_opt(asim, "debug",
				     "Debug messages appears on this console."));

  serial_in = (FILE*)asim->get_parg(0, "Ser_in");
  serial_out= (FILE*)asim->get_parg(0, "Ser_out");
  if (serial_in)
    {
      // making `serial' unbuffered
      if (setvbuf(serial_in, NULL, _IONBF, 0))
	perror("Unbuffer serial input channel");
      // setting O_NONBLOCK
      if ((i= fcntl(fileno(serial_in), F_GETFL, 0)) < 0)
	perror("Get flags of serial input");
      i|= O_NONBLOCK;
      if (fcntl(fileno(serial_in), F_SETFL, i) < 0)
	perror("Set flags of serial input");
      // switching terminal to noncanonical mode
      if (isatty(fileno(serial_in)))
	{
	  tcgetattr(fileno(serial_in), &saved_attributes_in);
	  tcgetattr(fileno(serial_in), &tattr);
	  tattr.c_lflag&= ~(ICANON|ECHO);
	  tattr.c_cc[VMIN] = 1;
	  tattr.c_cc[VTIME]= 0;
	  tcsetattr(fileno(serial_in), TCSAFLUSH, &tattr);
	}
      else
	fprintf(stderr, "Warning: serial input interface connected to a "
		"non-terminal file.\n");
    }
  if (serial_out)
    {
      // making `serial' unbuffered
      if (setvbuf(serial_out, NULL, _IONBF, 0))
	perror("Unbuffer serial output channel");
      // setting O_NONBLOCK
      if ((i= fcntl(fileno(serial_out), F_GETFL, 0)) < 0)
	perror("Get flags of serial output");
      i|= O_NONBLOCK;
      if (fcntl(fileno(serial_out), F_SETFL, i) < 0)
	perror("Set flags of serial output");
      // switching terminal to noncanonical mode
      if (isatty(fileno(serial_out)))
	{
	  tcgetattr(fileno(serial_out), &saved_attributes_out);
	  tcgetattr(fileno(serial_out), &tattr);
	  tattr.c_lflag&= ~(ICANON|ECHO);
	  tattr.c_cc[VMIN] = 1;
	  tattr.c_cc[VTIME]= 0;
	  tcsetattr(fileno(serial_out), TCSAFLUSH, &tattr);
	}
      else
	fprintf(/*cmd_out*/stderr, "Warning: serial output interface connected to a "
		"non-terminal file.\n");
    }

  memset(inst_map, '\0', sizeof(inst_map));
  for (i= 0; i < 4; i++)
    port_pins[i]= 0xff;
  it_sources= new cl_list(2, 2);
  it_sources->add(new cl_it_src(bmEX0, TCON, bmIE0, 0x0003, true,
				"external #0"));
  it_sources->add(new cl_it_src(bmET0, TCON, bmTF0, 0x000b, true,
				"timer #0"));
  it_sources->add(new cl_it_src(bmEX1, TCON, bmIE1, 0x0013, true,
				"external #1"));
  it_sources->add(new cl_it_src(bmET1, TCON, bmTF1, 0x001b, true,
				"timer #1"));
  it_sources->add(new cl_it_src(bmES , SCON, bmTI , 0x0023, false,
				"serial transmit"));
  it_sources->add(new cl_it_src(bmES , SCON, bmRI , 0x0023, false,
				"serial receive"));
  fbrk= new brk_coll(2, 2);
  ebrk= new brk_coll(2, 2);
  fbrk->Duplicates= FALSE;
}


/*
 * Initializing. Virtual calls go here
 * This method must be called first after object creation.
 */

int
t_uc51::init(void)
{
  int i;

  cl_uc::init();
  for (i= 0; i < IRAM_SIZE; i++)
    set_mem(MEM_IRAM, i, 0);
  for (i= 0; i < XRAM_SIZE; i++)
    set_mem(MEM_XRAM, i, 0);
  for (i= 0; i < EROM_SIZE; i++)
    set_mem(MEM_ROM, i, 0xff);
  reset();
}

static char id_string_51[100];

char *
t_uc51::id_string(void)
{
  int i;
  char *s= (char*)malloc(100);

  for (i= 0; cpus_51[i].type_str != NULL && cpus_51[i].type != type; i++) ;
  sprintf(id_string_51, "%s %s",
	  cpus_51[i].type_str?cpus_51[i].type_str:"51",
	  (technology==CPU_HMOS)?"HMOS":"CMOS");
  return(id_string_51);
}


/*
 * Destroying the micro-controller object
 */

t_uc51::~t_uc51(void)
{
  struct it_level *p;

  while (it_levels->next)
    {
      p= it_levels->next;
      free(it_levels);
      it_levels= p;
    }
  free(it_levels);
  delete it_sources;
  delete fbrk;
  delete ebrk;
  /*if (cmd_in &&
      cmd_in != stdin)
    fclose(cmd_in);
  if (cmd_out &&
      cmd_out != stdout)
      fclose(cmd_out);*/
  if (serial_out)
    {
      if (isatty(fileno(serial_out)))
	tcsetattr(fileno(serial_out), TCSANOW, &saved_attributes_out);
      fclose(serial_out);
    }
  if (serial_in)
    {
      if (isatty(fileno(serial_in)))
	tcsetattr(fileno(serial_in), TCSANOW, &saved_attributes_in);
      fclose(serial_in);
    }
}


/*
 * Writing data to EROM
 */

void
t_uc51::write_rom(uint addr, ulong data)
{
  if (addr < EROM_SIZE)
    write_mem(MEM_ROM, addr, data);
}


/*
 * Disassembling an instruction
 */

struct dis_entry *
t_uc51::dis_tbl(void)
{
  return(disass_51);
}

struct name_entry *
t_uc51::sfr_tbl(void)
{
  return(sfr_tab);
}

struct name_entry *
t_uc51::bit_tbl(void)
{
  return(bit_tab);
}

char *
t_uc51::disass(uint addr, char *sep)
{
  return(do_disass(dis_tbl(),
		   sfr_tab,
		   bit_tab,
		   addr, sep));
}

char *
t_uc51::do_disass(struct dis_entry *tabl,
		struct name_entry *sfr_tab,
		struct name_entry *bit_tab,
		uint addr, char *sep)
{
  char work[256], temp[20], c[2];
  char *buf, *p, *b, *t;

  p= work;
  b= tabl[get_mem(MEM_ROM, addr)].mnemonic;
  while (*b)
    {
      if (*b == '%')
	{
	  b++;
	  switch (*(b++))
	    {
	    case 'A': // absolute address
	      sprintf(temp, "%04x",
		      (addr&0xf800)|
		      (((get_mem(MEM_ROM, addr)>>5)&0x07)*256 +
		       get_mem(MEM_ROM, addr+1)));
	      break;
	    case 'l': // long address
	      sprintf(temp, "%04x",
		      get_mem(MEM_ROM, addr+1)*256 + get_mem(MEM_ROM, addr+2));
	      break;
	    case 'a': // addr8 (direct address) at 2nd byte
 	      if (!get_name(get_mem(MEM_ROM, addr+1), sfr_tab, temp))
		sprintf(temp, "%02x", get_mem(MEM_ROM, addr+1));
	      break;
	    case '8': // addr8 (direct address) at 3rd byte
 	      if (!get_name(get_mem(MEM_ROM, addr+2), sfr_tab, temp))
		sprintf(temp, "%02x", get_mem(MEM_ROM, addr+1));
	      sprintf(temp, "%02x", get_mem(MEM_ROM, addr+2));
	      break;
	    case 'b': // bitaddr at 2nd byte
	      if (get_name(get_mem(MEM_ROM, addr+1), bit_tab, temp))
		break;
	      if (get_name(get_bitidx(get_mem(MEM_ROM, addr+1)),
			   sfr_tab, temp))
		{
		  strcat(temp, ".");
		  sprintf(c, "%1d", get_mem(MEM_ROM, addr+1)&0x07);
		  strcat(temp, c);
		  break;
		}
	      sprintf(temp, "%02x.%d",
		      get_bitidx(get_mem(MEM_ROM, addr+1)),
		      get_mem(MEM_ROM, addr+1)&0x07);
	      break;
	    case 'r': // rel8 address at 2nd byte
	      sprintf(temp, "%04x",
		      addr+2+(signed char)(get_mem(MEM_ROM, addr+1)));
	      break;
	    case 'R': // rel8 address at 3rd byte
	      sprintf(temp, "%04x",
		      addr+3+(signed char)(get_mem(MEM_ROM, addr+2)));
	      break;
	    case 'd': // data8 at 2nd byte
	      sprintf(temp, "%02x", get_mem(MEM_ROM, addr+1));
	      break;
	    case 'D': // data8 at 3rd byte
	      sprintf(temp, "%02x", get_mem(MEM_ROM, addr+2));
	      break;
	    case '6': // data16 at 2nd(H)-3rd(L) byte
	      sprintf(temp, "%04x",
		      get_mem(MEM_ROM, addr+1)*256 + get_mem(MEM_ROM, addr+2));
	      break;
	    default:
	      strcpy(temp, "?");
	      break;
	    }
	  t= temp;
	  while (*t)
	    *(p++)= *(t++);
	}
      else
	*(p++)= *(b++);
    }
  *p= '\0';

  p= strchr(work, ' ');
  if (!p)
    {
      buf= strdup(work);
      return(buf);
    }
  if (sep == NULL)
    buf= (char *)malloc(6+strlen(p)+1);
  else
    buf= (char *)malloc((p-work)+strlen(sep)+strlen(p)+1);
  for (p= work, b= buf; *p != ' '; p++, b++)
    *b= *p;
  p++;
  *b= '\0';
  if (sep == NULL)
    {
      while (strlen(buf) < 6)
	strcat(buf, " ");
    }
  else
    strcat(buf, sep);
  strcat(buf, p);
  return(buf);
}


/*
 * Returning length of the instruction
 */

int
t_uc51::inst_length(uchar code)
{
  return(disass_51[code].length);
}


/*
 * Searching for a symbolic name in a name table
 */

bool
t_uc51::get_name(uchar addr, struct name_entry tab[], char *buf)
{
  int i;

  i= 0;
  while (tab[i].name &&
	 (!(tab[i].cpu_type & type) ||
	 (tab[i].addr != addr)))
    i++;
  if (tab[i].name)
    strcpy(buf, tab[i].name);
  return(tab[i].name != NULL);
}


/*
 * Resetting the micro-controller
 */

void
t_uc51::reset(void)
{
  struct it_level *p;

  clear_sfr();

  PC    = 0;
  result= resGO;
  state = stGO;

  was_reti= FALSE;
  while (it_levels->next)
    {
      p= it_levels->next;
      free(it_levels);
      it_levels= p;
    }

  s_tr_t1    = 0;
  s_rec_t1   = 0;
  s_tr_tick  = 0;
  s_rec_tick = 0;
  s_in       = 0;
  s_out      = 0;
  s_sending  = FALSE;
  s_receiving= FALSE;
  s_rec_bit  = 0;
  s_tr_bit   = 0;
}


/*
 * Setting up SFR area to reset value
 */

void
t_uc51::clear_sfr(void)
{
  int i;
  
  for (i= 0; i < SFR_SIZE; i++)
    set_mem(MEM_SFR, i, 0);
  set_mem(MEM_SFR, P0, 0xff);
  set_mem(MEM_SFR, P1, 0xff);
  set_mem(MEM_SFR, P2, 0xff);
  set_mem(MEM_SFR, P3, 0xff);
  set_mem(MEM_SFR, SP, 7);
  prev_p1= port_pins[1] & get_mem(MEM_SFR, P1);
  prev_p3= port_pins[3] & get_mem(MEM_SFR, P3);
}


/*
 * Analyzing code and settig up instruction map
 */

void
t_uc51::analyze(uint addr)
{
  do_analyze(disass_51, addr);
}


void
t_uc51::do_analyze(struct dis_entry *tabl, uint addr)
{
  while (!inst_at(addr) &&
	 get_mem(MEM_ROM, addr) != 0xa5 /* break point */)
    {
      set_inst_at(addr);
      switch (tabl[get_mem(MEM_ROM, addr)].branch)
	{
	case 'a': // acall
	  do_analyze(tabl,
		     (addr & 0xf800)|
		     ((get_mem(MEM_ROM, addr+1)&0x07)*256+
		      get_mem(MEM_ROM, addr+2)));
	  do_analyze(tabl, addr+tabl[get_mem(MEM_ROM, addr)].length);
	  break;
	case 'A': // ajmp
	  addr= (addr & 0xf800)|
	    ((get_mem(MEM_ROM, addr+1) & 0x07)*256 + get_mem(MEM_ROM, addr+2));
	  break;
	case 'l': // lcall
	  do_analyze(tabl,
		     get_mem(MEM_ROM, addr+1)*256 + get_mem(MEM_ROM, addr+2));
	  do_analyze(tabl, addr+tabl[get_mem(MEM_ROM, addr)].length);
	  break;
	case 'L': // ljmp
	  addr= get_mem(MEM_ROM, addr+1)*256 + get_mem(MEM_ROM, addr+2);
	  break;
	case 'r': // reljmp (2nd byte)
	  do_analyze(tabl,
		     (addr + (signed char)(get_mem(MEM_ROM, addr+1))) &
		     (EROM_SIZE - 1));
	  do_analyze(tabl, addr+tabl[get_mem(MEM_ROM, addr)].length);
	  break;
	case 'R': // reljmp (3rd byte)
	  do_analyze(tabl,
		     (addr+(signed char)
		      (get_mem(MEM_ROM, addr+2)))&(EROM_SIZE-1));
	  do_analyze(tabl, addr+tabl[get_mem(MEM_ROM, addr)].length);
	  break;
	case 's': // sjmp
	  addr= (addr+(signed char)((get_mem(MEM_ROM, addr+1))))&(EROM_SIZE-1);
	  break;
	case '_':
	  return;
	default:
	  addr= (addr+tabl[get_mem(MEM_ROM, addr)].length) & (EROM_SIZE - 1);
	  break;
	}
    }
}


/*
 * Handling instruction map
 *
 * `inst_at' is checking if the specified address is in instruction
 * map and `set_inst_at' marks the address in the map and
 * `del_inst_at' deletes the mark. `there_is_inst' cheks if there is
 * any mark in the map
 */

bool
t_uc51::inst_at(uint addr)
{
  return(inst_map[addr/8] & (1 << (addr & 0x07)));
}

void
t_uc51::set_inst_at(uint addr)
{
  inst_map[addr/8]|= (1 << (addr & 0x07));
}

void
t_uc51::del_inst_at(uint addr)
{
  inst_map[addr/8]&= ~(1 << (addr & 0x07));
}

bool
t_uc51::there_is_inst(void)
{
  uchar res= 0;
  int   i;
  
  for (i= 0; i < (EROM_SIZE/8); i++)
    res|= inst_map[i];
  return(res);
}


/*
 * Checking if there is a breakpoint at the specified address
 */

class cl_fetch_brk *
t_uc51::fbrk_at(uint addr)
{
  int idx;
  
  if (!fbrk->search(&addr, idx))
    return(0);
  return((class cl_fetch_brk *)(fbrk->at(idx)));
}


class cl_ev_brk *
t_uc51::ebrk_at(uint addr, char *id)
{
  int i;
  class cl_ev_brk *eb;

  for (i= 0; i < ebrk->count; i++)
    {
      eb= (class cl_ev_brk *)(ebrk->at(i));
      if (eb->addr == addr &&
	  !strcmp(eb->id, id))
	return(eb);
    }
  return(0);
}


/*
 * Deleting breakpoint from specified address
 */

void
t_uc51::rm_fbrk(uint addr)
{
  int idx;

  if (fbrk->search(&addr, idx))
    fbrk->free_at(idx);
}


void
t_uc51::rm_ebrk(uint addr, char *id)
{
  int i;
  class cl_ev_brk *eb;

  for (i= 0; i < ebrk->count; i++)
    {
      eb= (class cl_ev_brk *)(ebrk->at(i));
      if (eb->addr == addr &&
	  !strcmp(eb->id, id))
	ebrk->free_at(i);
    }
}


/*
 * Inform hardware elements that `cycles' machine cycles have elapsed
 */

int
t_uc51::tick(int cycles)
{
  cl_uc::tick(cycles);
  do_hardware(cycles);
  s_tr_tick+= cycles * clock_per_cycle();
  s_rec_tick+= cycles * clock_per_cycle();
}


/*
 * Correcting direct address
 *
 * This function returns address of addressed element which can be an IRAM
 * or an SFR.
 */

uchar *
t_uc51::get_direct(uchar addr)
{
  if (addr < SFR_START)
    return(&(MEM(MEM_IRAM)[addr]));
  else
    return(&(MEM(MEM_SFR)[addr]));
}

uchar *
t_uc51::get_direct(uchar addr, int *ev_i, int *ev_s)
{
  if (addr < SFR_START)
    return(&(MEM(MEM_IRAM)[*ev_i= addr]));
  else
    return(&(MEM(MEM_SFR)[*ev_s= addr]));
}


/*
 * Calculating address of indirectly addressed IRAM cell
 * If CPU is 8051 and addr is over 127, it must be illegal!
 */

uchar *
t_uc51::get_indirect(uchar addr, int *res)
{
  if (addr >= SFR_START)
    *res= resINV_ADDR;
  else
    *res= resGO;
  return(&(MEM(MEM_IRAM)[addr]));
}


/*
 * Calculating address of specified register cell in IRAM
 */

uchar *
t_uc51::get_reg(uchar regnum)
{
  return(&(MEM(MEM_IRAM)[(get_mem(MEM_SFR, PSW) & (bmRS0|bmRS1)) |
			(regnum & 0x07)]));
}

uchar *
t_uc51::get_reg(uchar regnum, int *event)
{
  return(&(MEM(MEM_IRAM)[*event= (get_mem(MEM_SFR, PSW) & (bmRS0|bmRS1)) |
			(regnum & 0x07)]));
}


/*
 * Calculating address of IRAM or SFR cell which contains addressed bit
 * Next function returns index of cell which contains addressed bit.
 */

uchar *
t_uc51::get_bit(uchar bitaddr)
{
  if (bitaddr < 128)
    return(&(MEM(MEM_IRAM)[(bitaddr/8)+32]));
  return(&(MEM(MEM_SFR)[bitaddr & 0xf8]));
}

uchar *
t_uc51::get_bit(uchar bitaddr, int *ev_i, int *ev_s)
{
  if (bitaddr < 128)
    return(&(MEM(MEM_IRAM)[*ev_i= (bitaddr/8)+32]));
  return(&(MEM(MEM_SFR)[*ev_s= bitaddr & 0xf8]));
}

uchar
t_uc51::get_bitidx(uchar bitaddr)
{
  if (bitaddr < 128)
    return((bitaddr/8)+32);
  return(bitaddr & 0xf8);
}


/*
 * Processing write operation to IRAM
 *
 * It starts serial transmition if address is in SFR and it is
 * SBUF. Effect on IE is also checked.
 */

void
t_uc51::proc_write(uchar *addr)
{
  if (addr == &(MEM(MEM_SFR)[SBUF]))
    {
      s_out= get_mem(MEM_SFR, SBUF);
      s_sending= TRUE;
      s_tr_bit = 0;
      s_tr_tick= 0;
      s_tr_t1  = 0;
    }
  if (addr == &(MEM(MEM_SFR)[IE]))
    was_reti= TRUE;
}


/*
 * Reading IRAM or SFR, but if address points to a port, it reads
 * port pins instead of port latches
 */

uchar
t_uc51::read(uchar *addr)
{
  if (addr == &(MEM(MEM_SFR)[P0]))
    return(get_mem(MEM_SFR, P0) & port_pins[0]);
  if (addr == &(MEM(MEM_SFR)[P1]))
    return(get_mem(MEM_SFR, P1) & port_pins[1]);
  if (addr == &(MEM(MEM_SFR)[P2]))
    return(get_mem(MEM_SFR, P2) & port_pins[2]);
  if (addr == &(MEM(MEM_SFR)[P3]))
    return(get_mem(MEM_SFR, P3) & port_pins[3]);
  return(*addr);
}


/*
 * Fetching one instruction and executing it
 */

void
t_uc51::pre_inst(void)
{}

int
t_uc51::exec_inst(bool use_breaks)
{
  uchar code;
  int res;

  pre_inst();
  code= fetch();
  tick(1);
  switch (code)
    {
    case 0x00:
      res= inst_nop(code);
      break;
    case 0x01: case 0x21: case 0x41: case 0x61:
    case 0x81: case 0xa1: case 0xc1: case 0xe1:
      res= inst_ajmp_addr(code);
      break;
    case 0x02:
      res= inst_ljmp(code);
      break;
    case 0x03:
      res= inst_rr(code);
      break;
    case 0x04:
      res= inst_inc_a(code);
      break;
    case 0x05:
      res= inst_inc_addr(code);
      break;
    case 0x06: case 0x07:
      res= inst_inc_$ri(code);
      break;
    case 0x08: case 0x09: case 0x0a: case 0x0b:
    case 0x0c: case 0x0d: case 0x0e: case 0x0f:
      res= inst_inc_rn(code);
      break;
    case 0x10:
      res= inst_jbc_bit_addr(code);
      break;
    case 0x11: case 0x31: case 0x51: case 0x71:
    case 0x91: case 0xb1: case 0xd1: case 0xf1:
      res= inst_acall_addr(code);
      break;
    case 0x12:
      res= inst_lcall(code, 0);
      break;
    case 0x13:
      res= inst_rrc(code);
      break;
    case 0x14:
      res= inst_dec_a(code);
      break;
    case 0x15:
      res= inst_dec_addr(code);
      break;
    case 0x16: case 0x17:
      res= inst_dec_$ri(code);
      break;
    case 0x18: case 0x19: case 0x1a: case 0x1b:
    case 0x1c: case 0x1d: case 0x1e: case 0x1f:
      res= inst_dec_rn(code);
      break;
    case 0x20:
      res= inst_jb_bit_addr(code);
      break;
    case 0x22:
      res= inst_ret(code);
      break;
    case 0x23:
      res= inst_rl(code);
      break;
    case 0x24:
      res= inst_add_a_$data(code);
      break;
    case 0x25:
      res= inst_add_a_addr(code);
      break;
    case 0x26: case 0x27:
      res= inst_add_a_$ri(code);
      break;
    case 0x28: case 0x29: case 0x2a: case 0x2b:
    case 0x2c: case 0x2d: case 0x2e: case 0x2f:
      res= inst_add_a_rn(code);
      break;
    case 0x30:
      res= inst_jnb_bit_addr(code);
      break;
    case 0x32:
      res= inst_reti(code);
      break;
    case 0x33:
      res= inst_rlc(code);
      break;
    case 0x34:
      res= inst_addc_a_$data(code);
      break;
    case 0x35:
      res= inst_addc_a_addr(code);
      break;
    case 0x36: case 0x37:
      res= inst_addc_a_$ri(code);
      break;
    case 0x38: case 0x39: case 0x3a: case 0x3b:
    case 0x3c: case 0x3d: case 0x3e: case 0x3f:
      res= inst_addc_a_rn(code);
      break;
    case 0x40:
      res= inst_jc_addr(code);
      break;
    case 0x42:
      res= inst_orl_addr_a(code);
      break;
    case 0x43:
      res= inst_orl_addr_$data(code);
      break;
    case 0x44:
      res= inst_orl_a_$data(code);
      break;
    case 0x45:
      res= inst_orl_a_addr(code);
      break;
    case 0x46: case 0x47:
      res= inst_orl_a_$ri(code);
      break;
    case 0x48: case 0x49: case 0x4a: case 0x4b:
    case 0x4c: case 0x4d: case 0x4e: case 0x4f:
      res= inst_orl_a_rn(code);
      break;
    case 0x50:
      res= inst_jnc_addr(code);
      break;
    case 0x52:
      res= inst_anl_addr_a(code);
      break;
    case 0x53:
      res= inst_anl_addr_$data(code);
      break;
    case 0x54:
      res= inst_anl_a_$data(code);
      break;
    case 0x55:
      res= inst_anl_a_addr(code);
      break;
    case 0x56: case 0x57:
      res= inst_anl_a_$ri(code);
      break;
    case 0x58: case 0x59: case 0x5a: case 0x5b:
    case 0x5c: case 0x5d: case 0x5e: case 0x5f:
      res= inst_anl_a_rn(code);
      break;
    case 0x60:
      res= inst_jz_addr(code);
      break;
    case 0x62:
      res= inst_xrl_addr_a(code);
      break;
    case 0x63:
      res= inst_xrl_addr_$data(code);
      break;
    case 0x64:
      res= inst_xrl_a_$data(code);
      break;
    case 0x65:
      res= inst_xrl_a_addr(code);
      break;
    case 0x66: case 0x67:
      res= inst_xrl_a_$ri(code);
      break;
    case 0x68: case 0x69: case 0x6a: case 0x6b:
    case 0x6c: case 0x6d: case 0x6e: case 0x6f:
      res= inst_xrl_a_rn(code);
      break;
    case 0x70:
      res= inst_jnz_addr(code);
      break;
    case 0x72:
      res= inst_orl_c_bit(code);
      break;
    case 0x73:
      res= inst_jmp_$a_dptr(code);
      break;
    case 0x74:
      res= inst_mov_a_$data(code);
      break;
    case 0x75:
      res= inst_mov_addr_$data(code);
      break;
    case 0x76: case 0x77:
      res= inst_mov_$ri_$data(code);
      break;
    case 0x78: case 0x79: case 0x7a: case 0x7b:
    case 0x7c: case 0x7d: case 0x7e: case 0x7f:
      res= inst_mov_rn_$data(code);
      break;
    case 0x80:
      res= inst_sjmp(code);
      break;
    case 0x82:
      res= inst_anl_c_bit(code);
      break;
    case 0x83:
      res= inst_movc_a_$a_pc(code);
      break;
    case 0x84:
      res= inst_div_ab(code);
      break;
    case 0x85:
      res= inst_mov_addr_addr(code);
      break;
    case 0x86: case 0x87:
      res= inst_mov_addr_$ri(code);
      break;
    case 0x88: case 0x89: case 0x8a: case 0x8b:
    case 0x8c: case 0x8d: case 0x8e: case 0x8f:
      res= inst_mov_addr_rn(code);
      break;
    case 0x90:
      res= inst_mov_dptr_$data(code);
      break;
    case 0x92:
      res= inst_mov_bit_c(code);
      break;
    case 0x93:
      res= inst_movc_a_$a_dptr(code);
      break;
    case 0x94:
      res= inst_subb_a_$data(code);
      break;
    case 0x95:
      res= inst_subb_a_addr(code);
      break;
    case 0x96: case 0x97:
      res= inst_subb_a_$ri(code);
      break;
    case 0x98: case 0x99: case 0x9a: case 0x9b:
    case 0x9c: case 0x9d: case 0x9e: case 0x9f:
      res= inst_subb_a_rn(code);
      break;
    case 0xa2:
      res= inst_mov_c_bit(code);
      break;
    case 0xa3:
      res= inst_inc_dptr(code);
      break;
    case 0xa4:
      res= inst_mul_ab(code);
      break;
    case 0xa5:
      if (use_breaks)
	res= break_point();
      else
	res= inst_unknown(code);
      break;
    case 0xa6: case 0xa7:
      res= inst_mov_$ri_addr(code);
      break;
    case 0xa8: case 0xa9: case 0xaa: case 0xab:
    case 0xac: case 0xad: case 0xae: case 0xaf:
      res= inst_mov_rn_addr(code);
      break;
    case 0xb0:
      res= inst_anl_c_$bit(code);
      break;
    case 0xb2:
      res= inst_cpl_bit(code);
      break;
    case 0xb3:
      res= inst_cpl_c(code);
      break;
    case 0xb4:
      res= inst_cjne_a_$data_addr(code);
      break;
    case 0xb5:
      res= inst_cjne_a_addr_addr(code);
      break;
    case 0xb6: case 0xb7:
      res= inst_cjne_$ri_$data_addr(code);
      break;
    case 0xb8: case 0xb9: case 0xba: case 0xbb:
    case 0xbc: case 0xbd: case 0xbe: case 0xbf:
      res= inst_cjne_rn_$data_addr(code);
      break;
    case 0xc0:
      res= inst_push(code);
      break;
    case 0xc2:
      res= inst_clr_bit(code);
      break;
    case 0xc3:
      res= inst_clr_c(code);
      break;
    case 0xc4:
      res= inst_swap(code);
      break;
    case 0xc5:
      res= inst_xch_a_addr(code);
      break;
    case 0xc6: case 0xc7:
      res= inst_xch_a_$ri(code);
      break;
    case 0xc8: case 0xc9: case 0xca: case 0xcb:
    case 0xcc: case 0xcd: case 0xce: case 0xcf:
      res= inst_xch_a_rn(code);
      break;
    case 0xd0:
      res= inst_pop(code);
      break;
    case 0xd2:
      res= inst_setb_bit(code);
      break;
    case 0xd3:
      res= inst_setb_c(code);
      break;
    case 0xd4:
      res= inst_da_a(code);
      break;
    case 0xd5:
      res= inst_djnz_addr_addr(code);
      break;
    case 0xd6: case 0xd7:
      res= inst_xchd_a_$ri(code);
      break;
    case 0xd8: case 0xd9: case 0xda: case 0xdb:
    case 0xdc: case 0xdd: case 0xde: case 0xdf:
      res= inst_djnz_rn_addr(code);
      break;
    case 0xe0:
      res= inst_movx_a_$dptr(code);
      break;
    case 0xe2: case 0xe3:
      res= inst_movx_a_$ri(code);
      break;
    case 0xe4:
      res= inst_clr_a(code);
      break;
    case 0xe5:
      res= inst_mov_a_addr(code);
      break;
    case 0xe6: case 0xe7:
      res= inst_mov_a_$ri(code);
      break;
    case 0xe8: case 0xe9: case 0xea: case 0xeb:
    case 0xec: case 0xed: case 0xee: case 0xef:
      res= inst_mov_a_rn(code);
      break;
    case 0xf0:
      res= inst_movx_$dptr_a(code);
      break;
    case 0xf2: case 0xf3:
      res= inst_movx_$ri_a(code);
      break;
    case 0xf4:
      res= inst_cpl_a(code);
      break;
    case 0xf5:
      res= inst_mov_addr_a(code);
      break;
    case 0xf6: case 0xf7:
      res= inst_mov_$ri_a(code);
      break;
    case 0xf8: case 0xf9: case 0xfa: case 0xfb:
    case 0xfc: case 0xfd: case 0xfe: case 0xff:
      res= inst_mov_rn_a(code);
      break;
    default:
      res= inst_unknown(code);
      break;
    }
  post_inst();
  return(res);
}

void
t_uc51::post_inst(void)
{}


/*
 * Simulating execution of next instruction
 *
 * This is an endless loop if requested number of steps is negative.
 * In this case execution is stopped if an instruction results other
 * status than GO. Execution can be stopped if `cmd_in' is not NULL
 * and there is input available on that file. It is usefull if the
 * command console is on a terminal. If input is available then a
 * complete line is read and dropped out because input is buffered
 * (inp_avail will be TRUE if ENTER is pressed) and it can confuse
 * command interepter.
 */

static bool
inp_avail(FILE *f)
{
  fd_set s;
  int i;
  struct timeval t;

  t.tv_sec= 0;
  t.tv_usec= 0;
  FD_ZERO(&s);
  FD_SET(fileno(f), &s);
  i= select(fileno(f)+1, &s, NULL, NULL, &t);
  return((i >= 0) && FD_ISSET(fileno(f), &s));
}

int
t_uc51::do_inst(int step)
{
  uchar code;
  bool  use_breaks;

  if (result == resHALT)
    return(result);

  use_breaks= step != 1;
  if (use_breaks)
    put_breaks();
  result= resGO;
  while ((result == resGO) &&
	 (state != stPD) &&
	 (step != 0))
    {
      step--;
      if (state == stGO)
	{
	  if (!inst_at(PC) &&
	      get_mem(MEM_ROM, PC) != 0xa5)
	    {
	      if (1)//debug)// && sim->cmd_out())
		{
		  /*f*/sim->cmd->debug(//sim->cmd_out(),
			  "Instruction (%02x) at %04x is not marked, "
			  "marking it...\n",
			  get_mem(MEM_ROM, PC), PC);
		  //fflush(sim->cmd_out());
		}
	      analyze(PC);
	    }
	  was_reti= FALSE;
	  result= exec_inst(use_breaks);
	  if (use_breaks &&
	      result == resGO)
	    result= check_events();
	}
      else
	tick(1);
      if (result == resGO)
	{
	  int res;
	  if ((res= do_hardware(0)) != resGO)
	    result= res;
	  else if ((res= do_interrupt()) != resGO)
	    result= res;
	  else
	    result= idle_pd();
	}
      if ((step < 0) &&
	  //(sim->cmd_in() != NULL) &&
	  //((Time % 1048576) > 600000) &&
	  /*inp_avail(sim->cmd_in())*/sim->cmd->input_avail())
	{
	  /*char c;
	    while ((c= getc(sim->cmd_in())) != '\n') ;*/
	  break;
	}
      if (((result == resINTERRUPT) &&
	   stop_at_it) ||
	  result == resHALT ||
	  result == resINV_ADDR ||
	  result == resSTACK_OV ||
	  result == resBREAKPOINT)
	{
	  sim->stop(result);
	  break;
	}
    }
  if (use_breaks)
    remove_breaks();
  return(result);
}


/*
 * Setting up parity flag
 */

void
t_uc51::set_p_flag(void)
{
  bool p;
  int i;
  uchar uc;

  p = FALSE;
  uc= get_mem(MEM_SFR, ACC);
  for (i= 0; i < 8; i++)
    {
      if (uc & 1)
	p= !p;
      uc>>= 1;
    }
  SET_BIT(p, PSW, bmP);
}

/*
 * Simulating hardware elements
 */

int
t_uc51::do_hardware(int cycles)
{
  int res_t, res_s;

  set_p_flag();

  // Read of SBUF must be serial input data
  set_mem(MEM_SFR, SBUF, s_in);

  res_t= do_timers(cycles);
  res_s= do_serial(cycles);

  // Setting up external interrupt request bits (IEx)
  if ((get_mem(MEM_SFR, TCON) & bmIT0))
    {
      // IE0 edge triggered
      if ((prev_p3 & bm_INT0) &&
	  !(get_mem(MEM_SFR, P3) & port_pins[3] & bm_INT0))
	// falling edge on INT0
	set_mem(MEM_SFR, TCON, get_mem(MEM_SFR, TCON)|bmIE0);
    }
  else
    {
      // IE0 level triggered
      if (get_mem(MEM_SFR, P3) & port_pins[3] & bm_INT0)
	set_mem(MEM_SFR, TCON, get_mem(MEM_SFR, TCON)|bmIE0);
      else
	set_mem(MEM_SFR, TCON, get_mem(MEM_SFR, TCON)&(~bmIE0));
    }
  if ((get_mem(MEM_SFR, TCON) & bmIT1))
    {
      // IE1 edge triggered
      if ((prev_p3 & bm_INT1) &&
	  !(get_mem(MEM_SFR, P3) & port_pins[3] & bm_INT1))
	// falling edge on INT1
	write_mem(MEM_SFR, TCON, get_mem(MEM_SFR, TCON)|bmIE1);
    }
  else
    {
      // IE1 level triggered
      if (get_mem(MEM_SFR, P3) & port_pins[3] & bm_INT1)
	set_mem(MEM_SFR, TCON, get_mem(MEM_SFR, TCON)|bmIE1);
      else
	set_mem(MEM_SFR, TCON, get_mem(MEM_SFR, TCON)&(~bmIE1));
    }
  prev_p3= get_mem(MEM_SFR, P3) & port_pins[3];
  prev_p1= get_mem(MEM_SFR, P3) & port_pins[1];

  res_t= do_wdt(cycles);

  if (res_t != resGO)
    return(res_t);
  return(res_s);
}


/*
 *
 */

int
t_uc51::serial_bit_cnt(void)
{
  int mode, divby;
  int *tr_src, *rec_src;

  mode= get_mem(MEM_SFR, SCON) >> 6;
  switch (mode)
    {
    case 0:
      divby  = 12;
      tr_src = &s_tr_tick;
      rec_src= &s_rec_tick;
      break;
    case 1:
    case 3:
      divby  = (get_mem(MEM_SFR, PCON)&bmSMOD)?16:32;
      tr_src = &s_tr_t1;
      rec_src= &s_rec_t1;
      break;
    case 2:
      divby  = (get_mem(MEM_SFR, PCON)&bmSMOD)?16:32;
      tr_src = &s_tr_tick;
      rec_src= &s_rec_tick;
      break;
    }
  if (s_sending)
    {
      while (*tr_src >= divby)
	{
	  (*tr_src)-= divby;
	  s_tr_bit++;
	}
    }
  if (s_receiving)
    {
      while (*rec_src >= divby)
	{
	  (*rec_src)-= divby;
	  s_rec_bit++;
	}
    }
}


/*
 * Simulating serial line
 */

static prev_bit_cnt= 0;

int
t_uc51::do_serial(int cycles)
{
  int mode, bits;
  char c; int i;

  mode= get_mem(MEM_SFR, SCON) >> 6;
  switch (mode)
    {
    case 0:
      bits= 8;
      break;
    case 1:
      bits= 10;
      break;
    case 2:
    case 3:
      bits= 11;
      break;
    }
  serial_bit_cnt();
  if (s_sending &&
      (s_tr_bit >= bits))
    {
      s_sending= FALSE;
      set_mem(MEM_SFR, SCON, get_mem(MEM_SFR, SCON)|bmTI);
      if (serial_out)
	{
	  fprintf(serial_out, "%c", (s_out=='\r')?'\n':s_out);
	  fflush(serial_out);
	}
      s_tr_bit-= bits;
    }
  if ((get_mem(MEM_SFR, SCON) & bmREN) &&
      serial_in &&
      !s_receiving)
    {
      fd_set set; static struct timeval timeout= {0,0};
      FD_ZERO(&set);
      FD_SET(fileno(serial_in), &set);
      i= select(fileno(serial_in)+1/*FD_SETSIZE*/,
		&set, NULL, NULL, &timeout);
      /*if (i>0 && ::read(fileno(serial_in), &c, 1) == 1)
	{
	  s_in= c;
	  s_receiving= TRUE;
	  }*/
      if (i>0)
	{
	  s_receiving= TRUE;
	  s_rec_bit= 0;
	  s_rec_tick= s_rec_t1= 0;
	}
    }
  if (s_receiving &&
      (s_rec_bit >= bits))
    {
      if (::read(fileno(serial_in), &c, 1) == 1)
	{
	  s_in= c;
	  set_mem(MEM_SFR, SBUF, s_in);
	}
      s_receiving= FALSE;
      received(c);
      s_rec_bit-= bits;
    }
  return(resGO);
}

void
t_uc51::received(int c)
{
  set_mem(MEM_SFR, SCON, get_mem(MEM_SFR, SCON)|bmRI);
}


/*
 * Simulating timers
 */

int
t_uc51::do_timers(int cycles)
{
  int res;

  if ((res= do_timer0(cycles)) != resGO)
    return(res);
  return(do_timer1(cycles));
}


/*
 * Simulating timer 0
 */

int
t_uc51::do_timer0(int cycles)
{
  if (((get_mem(MEM_SFR, TMOD) & bmGATE0) &&
      (get_mem(MEM_SFR, P3) & port_pins[3] & bm_INT0)) ||
      (get_mem(MEM_SFR, TCON) & bmTR0))
    {
      if (!(get_mem(MEM_SFR, TMOD) & bmC_T0) ||
	  ((prev_p3 & bmT0) &&
	   !(get_mem(MEM_SFR, P3) & port_pins[3] & bmT0)))
	{
	  if (!(get_mem(MEM_SFR, TMOD) & bmM00) &&
	      !(get_mem(MEM_SFR, TMOD) & bmM10))
	    {
	      if (get_mem(MEM_SFR, TMOD) & bmC_T0)
		cycles= 1;
	      while (cycles--)
		{
		  // mod 0, TH= 8 bit t/c, TL= 5 bit precounter
		  set_mem(MEM_SFR, TL0, get_mem(MEM_SFR, TL0)+1);
		  if (get_mem(MEM_SFR, TL0) > 0x1f)
		    {
		      set_mem(MEM_SFR, TL0, get_mem(MEM_SFR, TL0)&0x1f);
		      if (!++(MEM(MEM_SFR)[TH0]))
			{
			  set_mem(MEM_SFR, TCON, get_mem(MEM_SFR, TCON)|bmTF0);
			  t0_overflow();
			}
		    }
		}
	    }
	  else if ((get_mem(MEM_SFR, TMOD) & bmM00) &&
		   !(get_mem(MEM_SFR, TMOD) & bmM10))
	    {
	      if (get_mem(MEM_SFR, TMOD) & bmC_T0)
		cycles= 1;
	      while (cycles--)
		{
		  // mod 1 TH+TL= 16 bit t/c
		  set_mem(MEM_SFR, TL0, get_mem(MEM_SFR, TL0)+1);
		  if (!(get_mem(MEM_SFR, TL0)))
		    {
		      set_mem(MEM_SFR, TH0, get_mem(MEM_SFR, TH0)+1);
		      if (!(get_mem(MEM_SFR, TH0)))
			{
			  set_mem(MEM_SFR, TCON, get_mem(MEM_SFR, TCON)|bmTF0);
			  t0_overflow();
			}
		    }
		}
	    }
	  else if (!(get_mem(MEM_SFR, TMOD) & bmM00) &&
		   (get_mem(MEM_SFR, TMOD) & bmM10))
	    {
	      if (get_mem(MEM_SFR, TMOD) & bmC_T0)
		cycles= 1;
	      while (cycles--)
		{
		  // mod 2 TL= 8 bit t/c auto reload from TH
		  set_mem(MEM_SFR, TL0, get_mem(MEM_SFR, TL0)+1);
		  if (!(get_mem(MEM_SFR, TL0)))
		    {
		      set_mem(MEM_SFR, TL0, get_mem(MEM_SFR, TH0));
		      MEM(MEM_SFR)[TCON]|= bmTF0;
		      t0_overflow();
		    }
		}
	    }
	  else
	    {
	      // mod 3 TL= 8 bit t/c
	      //       TH= 8 bit timer controlled with T1's bits
	      if (!++(MEM(MEM_SFR)[TL0]))
		{
		  MEM(MEM_SFR)[TCON]|= bmTF0;
		  t0_overflow();
		}
	    }
	}
    }
  if ((MEM(MEM_SFR)[TMOD] & bmM00) &&
      (MEM(MEM_SFR)[TMOD] & bmM10))
    {
      if (((MEM(MEM_SFR)[TMOD] & bmGATE1) &&
	   (MEM(MEM_SFR)[P3] & port_pins[3] & bm_INT1)) ||
	  (MEM(MEM_SFR)[TCON] & bmTR1))
	{
	  if (!++(MEM(MEM_SFR)[TH0]))
	    {
	      MEM(MEM_SFR)[TCON]|= bmTF1;
	      s_tr_t1++;
	      s_rec_t1++;
	      t0_overflow();
	    }
	}
    }
  return(resGO);
}

/*
 * Called every time when T0 overflows
 */

int
t_uc51::t0_overflow(void)
{}


/*
 * Simulating timer 1
 */

int
t_uc51::do_timer1(int cycles)
{
  if (((get_mem(MEM_SFR, TMOD) & bmGATE1) &&
      (get_mem(MEM_SFR, P3) & port_pins[3] & bm_INT1)) ||
      (get_mem(MEM_SFR, TCON) & bmTR1))
    {
      if (!(get_mem(MEM_SFR, TMOD) & bmC_T1) ||
	  ((prev_p3 & bmT1) &&
	   !(get_mem(MEM_SFR, P3) & port_pins[3] & bmT1)))
	{
	  if (!(get_mem(MEM_SFR, TMOD) & bmM01) &&
	      !(get_mem(MEM_SFR, TMOD) & bmM11))
	    {
	      if (get_mem(MEM_SFR, TMOD) & bmC_T0)
		cycles= 1;
	      while (cycles--)
		{
		  // mod 0, TH= 8 bit t/c, TL= 5 bit precounter
		  if (++(MEM(MEM_SFR)[TL1]) > 0x1f)
		    {
		      MEM(MEM_SFR)[TL1]&= 0x1f;
		      if (!++(MEM(MEM_SFR)[TH1]))
			{
			  MEM(MEM_SFR)[TCON]|= bmTF1;
			  s_tr_t1++;
			  s_rec_t1++;
			}
		    }
		}
	    }
	  else if ((MEM(MEM_SFR)[TMOD] & bmM01) &&
		   !(MEM(MEM_SFR)[TMOD] & bmM11))
	    {
	      if (MEM(MEM_SFR)[TMOD] & bmC_T0)
		cycles= 1;
	      while (cycles--)
		{
		  // mod 1 TH+TL= 16 bit t/c
		  if (!++(MEM(MEM_SFR)[TL1]))
		    if (!++(MEM(MEM_SFR)[TH1]))
		      {
			MEM(MEM_SFR)[TCON]|= bmTF1;
			s_tr_t1++;
			s_rec_t1++;
		      }
		}
	    }
	  else if (!(MEM(MEM_SFR)[TMOD] & bmM01) &&
		   (MEM(MEM_SFR)[TMOD] & bmM11))
	    {
	      if (MEM(MEM_SFR)[TMOD] & bmC_T0)
		cycles= 1;
	      while (cycles--)
		{
		  // mod 2 TL= 8 bit t/c auto reload from TH
		  if (!++(MEM(MEM_SFR)[TL1]))
		    {
		      MEM(MEM_SFR)[TL1]= MEM(MEM_SFR)[TH1];
		      MEM(MEM_SFR)[TCON]|= bmTF1;
		      s_tr_t1++;
		      s_rec_t1++;
		    }
		}
	    }
	  else
	    // mod 3 stop
	    ;
	}
    }
  return(resGO);
}


/*
 * Abstract method to handle WDT
 */

int
t_uc51::do_wdt(int cycles)
{
  return(resGO);
}


/*
 * Checking for interrupt requests and accept one if needed
 */

int
t_uc51::do_interrupt(void)
{
  //uchar mask;
  class cl_it_src *it;
  int i, pr;

  if (was_reti)
    {
      was_reti= FALSE;
      return(resGO);
    }
  if (!(get_mem(MEM_SFR, IE) & bmEA))
    return(resGO);
  /*if (it_levels->level == IT_HIGH)
    return(resGO);
  if (it_levels->level == IT_LOW)
    mask= MEM(MEM_SFR)[IP];
  else
  mask= 0xff;*/
  for (i= 0; i < it_sources->count; i++)
    {
      it= (class cl_it_src *)(it_sources->at(i));
      if (it->is_active() &&
	  (get_mem(MEM_SFR, IE) & it->ie_mask) &&
	  (get_mem(MEM_SFR, it->src_reg) & it->src_mask))
	{
	  if ((pr= it_priority(it->ie_mask)) <= it_levels->level)
	    continue;
	  if (it->clr_bit)
	    MEM(MEM_SFR)[it->src_reg]&= ~(it->src_mask);
	  sim->cmd->debug("%g sec: Accepting interrupt `%s' PC= %06x\n",
			  get_rtime(), it->name, PC);
	  return(accept_it(pr, it->addr));
	}
    }
  return(resGO);
}

int
t_uc51::it_priority(uchar ie_mask)
{
  if (MEM(MEM_SFR)[IP] & ie_mask)
    return(1);
  return(0);
}


/*
 * Accept an interrupt
 */

int
t_uc51::accept_it(int level, uint addr)
{
  struct it_level *p;
  int res;

  state= stGO;
  write_mem(MEM_SFR, PCON, get_mem(MEM_SFR, PCON)&(~bmIDL));
  p= (struct it_level *)malloc(sizeof(struct it_level));
  p->level= level;
  p->next = it_levels;
  it_levels= p;
  tick(1);
  if ((res= inst_lcall(0, addr)) != resGO)
    return(res);
  else
    return(resINTERRUPT);
}


/*
 * Checking if Idle or PowerDown mode should be activated
 */

int
t_uc51::idle_pd(void)
{
  if (technology != CPU_CMOS)
    return(resGO);
  if (get_mem(MEM_SFR, PCON) & bmIDL)
    state= stIDLE;
  if (get_mem(MEM_SFR, PCON) & bmPD)
    state= stPD;
  return(resGO);
}


/*
 * Inserting FETCH breakpoints into the code
 */

void
t_uc51::put_breaks(void)
{
  class cl_fetch_brk *b;
  int   i;

  for (i= 0; i < fbrk->count; i++)
    {
      b= (class cl_fetch_brk *)(fbrk->at(i));
      if (!inst_at(b->addr) &&
	  debug)
	sim->cmd->printf(
		"Warning: maybe not instruction at breakpoint %04x\n",
		b->addr);
      b->code= get_mem(MEM_ROM, b->addr);
      set_mem(MEM_ROM, b->addr, 0xa5);
    }
  event_at.wi= -1;
  event_at.ri= -1;
  event_at.wx= -1;
  event_at.rx= -1;
  event_at.ws= -1;
  event_at.rs= -1;
  event_at.rc= -1;
}


/*
 * Removing FETCH breakpoints from the code and restore original values
 */

void
t_uc51::remove_breaks(void)
{
  class cl_fetch_brk *b;
  int   i;

  for (i= 0; i < fbrk->count; i++)
    {
      b= (class cl_fetch_brk *)(fbrk->at(i));
      set_mem(MEM_ROM, b->addr, b->code);
    }
}


/*
 * Checking if EVENT break happened
 */

int
t_uc51::check_events(void)
{
  int i;
  class cl_ev_brk *eb;

  if (!ebrk->count)
    return(resGO);
  for (i= 0; i < ebrk->count; i++)
    {
      eb= (class cl_ev_brk *)(ebrk->at(i));
      if (eb->match(&event_at))
	return(resBREAKPOINT);
    }
  return(resGO);
}


/*
 * Simulating an unknown instruction
 *
 * Normally this function is called for unimplemented instructions, because
 * every instruction must be known!
 */

int
t_uc51::inst_unknown(uchar code)
{
  PC--;
  if (1)//debug)// && sim->cmd_out())
    sim->cmd->debug("Unknown instruction %02x at %06x\n", code, PC);
  return(resHALT);
}


/*
 * 0x00 1 12 NOP
 */

int
t_uc51::inst_nop(uchar code)
{
  return(resGO);
}


/*
 * 0xe4 1 12 CLR A
 */

int
t_uc51::inst_clr_a(uchar code)
{
  write_mem(MEM_SFR, ACC, 0);
  return(resGO);
}


/*
 * 0xc4 1 1 SWAP A
 */

int
t_uc51::inst_swap(uchar code)
{
  uchar temp;

  temp= (MEM(MEM_SFR)[ACC] >> 4) & 0x0f;
  MEM(MEM_SFR)[ACC]= (MEM(MEM_SFR)[ACC] << 4) | temp;
  return(resGO);
}


/*
 * Breakpoint
 */

int
t_uc51::break_point(void)
{
  class cl_fetch_brk *fb;
  int   idx;

  PC--;
  if (!fbrk->search(&PC, idx))
    {
      PC++;
      return(inst_unknown(0xa5));
    }
  fb= (class cl_fetch_brk *)(fbrk->at(idx));
  if (--(fb->cnt) == 0)
    {
      fb->cnt= fb->hit;
      if (fb->perm == brkDYNAMIC)
	{
	  set_mem(MEM_ROM, fb->addr, fb->code);
	  fbrk->free_at(idx);
	}
      return(resBREAKPOINT);
    }
  return(resGO);
}


/* End of uc51.cc */
